{ AI Unit Designed by Jason McMillen  2002-2003 }
unit AIUnit;

interface

uses Math, Dialogs, SysUtils;

type
  TNeuron = class(TObject)
    DigitalOutput: Boolean;
    NumberOfInputs: Integer;
    Weight: Array of Real;{Double}
    constructor Init(NumOfInputs: Integer; Digital: Boolean);
    function Output(Inputs: Array of Real): Real;
  end;
  TNeuronLayer = class(TObject)
    NumberOfNeurons: Integer;
    Neurons: Array of TNeuron;
    constructor Init(NumOfNeurons, NumOfInputsPerNeuron: Integer; Digital: Boolean);
    procedure Execute(Inputs: Array of Real; var Outputs: Array of Real);
  end;
  TNeuralNet = class(TObject)
   private
    NumberOfInputs: Integer;
    NumberOfOutputs: Integer;
    NumberOfLayers: Integer;
    //storage for each layer of neurons including the output layer
    Layers: Array of TNeuronLayer;
   public
    FitnessScore: Integer;
    DigitalOutput: Boolean;
    constructor Init(NumOfInputs: Integer; NeuronSet: Array of Integer; Digital: Boolean);
    destructor Kill;
    function GetNumberOfWeights: Integer;                 //returns the total number of weights in the net
    procedure GetWeights(var Weights: Array of Real);     //gets the weights from the NN
    procedure PutWeights(Weights: Array of Real);         //replaces the weights with new ones
//    function Sigmoid(Activation, Response: Real): Real;   //sigmoid response curve
    procedure Execute(Input: Array of Real; var Output: Array of Real);//calculates the outputs from a set of inputs
  end;
  TChromosome = class(TObject)
    NumberOfValues: Integer;
    Values: Array of Real;{Double}
    FitnessScore: Integer;
    constructor Init(Size: Integer);
    destructor DeInit;
    procedure GetChromo(var Chromo: Array of Real);
    procedure PutChromo(Chromo: Array of Real);
  end;

const
  CrossOver_Rate = 0.7;
  Mutation_Rate = 0.05;{0.2 to 0.05 recommended}
  Digital_Mutation_Rate = 0.001;
  ThresholdCheck_On = True;

function Sigmoid(Activation, Response: Real): Real;

procedure GetChromosomes(NeuralNets: Array of TNeuralNet; var Population: Array of TChromosome; ChromoSize: Integer);
procedure Apply_Basic_Algorithm(var Population: Array of TChromosome; NumberOfPopulation, ChromoSize: Integer);
procedure PutChromosomes(var NeuralNets: Array of TNeuralNet; Population: Array of TChromosome; ChromoSize: Integer);

procedure Apply_SpawnStrongest_Algorithm(var Population: Array of TChromosome; NumberOfPopulation, ChromoSize: Integer);
procedure Apply_SpawnNumberOfHighest_Algorithm(var Population: Array of TChromosome; NumberOfPopulation, ChromoSize, EvolutionCount: Integer);

implementation

function Sigmoid(Activation, Response: Real): Real;
const e = 2.7183;
begin
     Result := 1 / (1 + Power(e, (-Activation / Response)));
end;

{-- Genetic Algorithm Functions --}
procedure GetChromosomes(NeuralNets: Array of TNeuralNet; var Population: Array of TChromosome; ChromoSize: Integer);
var i: Integer;
    TempValues: Array of Real;
begin
     if (Length(NeuralNets) > 0) then
        if (Length(NeuralNets) = Length(Population)) then
           for i := 0 to Length(NeuralNets) - 1 do
           begin
                Population[i].FitnessScore := NeuralNets[i].FitnessScore;
                SetLength(TempValues, ChromoSize);
                NeuralNets[i].GetWeights(TempValues);
                Population[i].PutChromo(TempValues);
           end
        else
        begin
             ShowMessage('ERROR <GetChromosomes()>: Non-Equal Array Sizes!');
             Halt;
        end;

     TempValues := nil;
end;
procedure PutChromosomes(var NeuralNets: Array of TNeuralNet; Population: Array of TChromosome; ChromoSize: Integer);
var i: Integer;
begin
     if (Length(NeuralNets) > 0) then
        if (Length(NeuralNets) = Length(Population)) then
           for i := 0 to Length(NeuralNets) - 1 do
           begin
                NeuralNets[i].FitnessScore := Population[i].FitnessScore;
                NeuralNets[i].PutWeights(Population[i].Values);
           end
        else
        begin
             ShowMessage('ERROR <PutChromosomes()>: Non-Equal Array Sizes!');
             Halt;
        end;
end;
procedure Apply_Basic_Algorithm(var Population: Array of TChromosome; NumberOfPopulation, ChromoSize: Integer);
var i, j, k: Integer;
    NewPopulation: Array of TChromosome;
    TotalFitness: Integer;
    RandomSelectionA, RandomSelectionB: Real;
    SelectionA, SelectionB: Array of Real;
    AFitness, BFitness: Integer;
    SplitPos: Integer;
begin
     {- Initalizing NewPopulation and getting Total Fitness -}
     SetLength(NewPopulation, Length(Population));
     TotalFitness := 0;
     for i := 0 to NumberOfPopulation - 1 do
     begin
          NewPopulation[i] := TChromosome.Init(ChromoSize);

          inc(Population[i].FitnessScore);{Every creature has 1 peice of the pie!}

          inc(TotalFitness, Population[i].FitnessScore);
     end;

     SetLength(SelectionA, ChromoSize);//Another Address problem!!!
     SetLength(SelectionB, ChromoSize);
     for i := 0 to NumberOfPopulation - 1 do
     begin
     {- Roulette Wheel Selection -}
          {Select A}
          RandomSelectionA := Random * TotalFitness;
          j := 0;
          repeat
          begin
               if (j > NumberOfPopulation - 1) then
                  j := 0;
               RandomSelectionA := RandomSelectionA - Population[j].FitnessScore;
               inc(j);
          end;
          until (RandomSelectionA < Population[j - 1].FitnessScore);

          for k := 0 to Population[j - 1].NumberOfValues - 1 do
              SelectionA[k] := Population[j - 1].Values[k];
          AFitness := Population[j - 1].FitnessScore;

          {Select B}
          RandomSelectionB := Random * TotalFitness; // !!!Remove Rounding!!!
          j := 0;
          repeat
          begin
               if (j > NumberOfPopulation - 1) then
                  j := 0;
               RandomSelectionB := RandomSelectionB - Population[j].FitnessScore;
               inc(j);
          end;
          until (RandomSelectionB < Population[j - 1].FitnessScore);

          for k := 0 to Population[j - 1].NumberOfValues - 1 do
              SelectionB[k] := Population[j - 1].Values[k];
          BFitness := Population[j - 1].FitnessScore;

     {- Cross-Over Rate -}
          if (Random <= CrossOver_Rate) then
          begin
               SplitPos := Round(Random * Population[i].NumberOfValues);
               {First Half from A}
               for j := 0 to SplitPos - 1 do
                   NewPopulation[i].Values[j] := SelectionA[j];
               {Second Half from B}
               for j := SplitPos to Population[i].NumberOfValues - 1 do
                   NewPopulation[i].Values[j] := SelectionB[j];
          end
          else
          begin
               if (AFitness > BFitness) then
                  for j := 0 to NewPopulation[i].NumberOfValues - 1 do
                      NewPopulation[i].Values[j] := SelectionA[j]
               else
                   for j := 0 to NewPopulation[i].NumberOfValues - 1 do
                       NewPopulation[i].Values[j] := SelectionB[j];
          end;


     {- Mutation Rate -}
          for j := 0 to NewPopulation[i].NumberOfValues - 1 do
              if (Random <= Mutation_Rate) then
                 NewPopulation[i].Values[j] := (Random * 2) - 1;
//                 NewPopulation[i].Values[j] + ((Random * 2) - 1); // Unlimit Bounds...
     end;


     {- Update Chromosomes -}
     for i := 0 to NumberOfPopulation - 1 do
         for j := 0 to Population[i].NumberOfValues - 1 do
         begin
              Population[i].Values[j] := NewPopulation[i].Values[j];
              Population[i].FitnessScore := 0;
         end;

     {- Destroy Temporary Objects -}
     for i := 0 to NumberOfPopulation - 1 do
         NewPopulation[i].DeInit;
     NewPopulation := nil;
     SelectionA := nil;
     SelectionB := nil;
end;
procedure Apply_SpawnStrongest_Algorithm(var Population: Array of TChromosome; NumberOfPopulation, ChromoSize: Integer);
var i, j: Integer;
    NewPopulation: Array of TChromosome;
//    TotalFitness : Integer;
    StrongestMember: Integer;
    StrongestFitness: Integer;
begin
     {- Initalizing NewPopulation and getting Total Fitness -}
     SetLength(NewPopulation, Length(Population));
//     TotalFitness := 0;
     for i := 0 to NumberOfPopulation - 1 do
     begin
          NewPopulation[i] := TChromosome.Init(ChromoSize);

          inc(Population[i].FitnessScore);{Every creature has 1 peice of the pie!}

//          inc(TotalFitness, Population[i].FitnessScore);
     end;


     StrongestMember := 0;
     StrongestFitness := Population[StrongestMember].FitnessScore;
     for i := 1 to NumberOfPopulation - 1 do
         if (Population[i].FitnessScore > StrongestFitness) then
            begin
                 StrongestMember := i;
                 StrongestFitness := Population[StrongestMember].FitnessScore;
            end;

     for i := 0 to NumberOfPopulation - 1 do
     begin
          for j := 0 to NewPopulation[i].NumberOfValues - 1 do
              NewPopulation[i].Values[j] := Population[StrongestMember].Values[j];

          {- Mutation Rate -}
          for j := 0 to NewPopulation[i].NumberOfValues - 1 do
              if (Random <= Mutation_Rate) then
                 NewPopulation[i].Values[j] := (Random * 2) - 1;
//                 NewPopulation[i].Values[j] + ((Random * 2) - 1); // Unlimit Bounds...
     end;


     {- Update Chromosomes -}
     for i := 0 to NumberOfPopulation - 1 do
         for j := 0 to Population[i].NumberOfValues - 1 do
         begin
              Population[i].Values[j] := NewPopulation[i].Values[j];
              Population[i].FitnessScore := 0;
         end;

     {- Destroy Temporary Objects -}
     for i := 0 to NumberOfPopulation - 1 do
         NewPopulation[i].DeInit;
     NewPopulation := nil;
end;
procedure Apply_SpawnNumberOfHighest_Algorithm(var Population: Array of TChromosome; NumberOfPopulation, ChromoSize, EvolutionCount: Integer);
var i, j, k: Integer;
    best_j, best_value, best_index: Integer;
    NewPopulation: Array of TChromosome;
    TotalFitness: Integer;
    RandomSelectionA, RandomSelectionB: Real;
    SelectionA, SelectionB: Array of Real;
    AFitness, BFitness: Integer;
    SplitPos: Integer;
    NewOrder: Array of Integer;
begin
     {- Initalizing NewPopulation and getting Total Fitness -}
     SetLength(NewPopulation, Length(Population));
     SetLength(NewOrder, Length(Population));
     TotalFitness := 0;
     for i := 0 to NumberOfPopulation - 1 do
     begin
          NewPopulation[i] := TChromosome.Init(ChromoSize);

          inc(TotalFitness, Population[i].FitnessScore);

          NewOrder[i] := i;
     end;

     { Select Sort Population Using NewOrder to Reorder Population by Fitness }
     {highest fitness = highest}
     for i := 0 to NumberOfPopulation - 2 do
     begin
          best_value := Population[NewOrder[i]].FitnessScore;
          best_index := NewOrder[i];
          best_j := i;
          for j := i + 1 to NumberOfPopulation - 1 do
          begin
               if (Population[NewOrder[j]].FitnessScore < best_value) Then
               begin
                    best_value := Population[NewOrder[j]].FitnessScore;
                    best_index := NewOrder[j];
                    best_j := j;
               end;
          end;    // for j := i + 1 to max do
//          Population[best_j].FitnessScore := Population[NewOrder[i]].FitnessScore;
          NewOrder[best_j] := i;
//          Population[NewOrder[i]].FitnessScore := best_value;
          NewOrder[i] := best_index;
     end;        // for i := min to max - 1 do
     { End of Sort Algo }

     {Eliminate Everyone but the top EvolutionCount}
     for i := 0 to NumberOfPopulation - 1 - EvolutionCount do
         Population[NewOrder[i]].FitnessScore := 0;

     SetLength(SelectionA, ChromoSize);//Another Address problem!!!
     SetLength(SelectionB, ChromoSize);
     for i := 0 to NumberOfPopulation - 1 do
     begin
     {- Roulette Wheel Selection -}
          {Select A}
          RandomSelectionA := Random * TotalFitness;
          j := 0;
          repeat
          begin
               if (j > NumberOfPopulation - 1) then
                  j := 0;
               RandomSelectionA := RandomSelectionA - Population[NewOrder[j]].FitnessScore;
               inc(j);
          end;
          until (RandomSelectionA < Population[NewOrder[j - 1]].FitnessScore);

          for k := 0 to Population[NewOrder[j - 1]].NumberOfValues - 1 do
              SelectionA[k] := Population[NewOrder[j - 1]].Values[k];
          AFitness := Population[NewOrder[j - 1]].FitnessScore;

          {Select B}
          RandomSelectionB := Random * TotalFitness; // !!!Remove Rounding!!!
          j := 0;
          repeat
          begin
               if (j > NumberOfPopulation - 1) then
                  j := 0;
               RandomSelectionB := RandomSelectionB - Population[NewOrder[j]].FitnessScore;
               inc(j);
          end;
          until (RandomSelectionB < Population[NewOrder[j - 1]].FitnessScore);

          for k := 0 to Population[NewOrder[j - 1]].NumberOfValues - 1 do
              SelectionB[k] := Population[NewOrder[j - 1]].Values[k];
          BFitness := Population[NewOrder[j - 1]].FitnessScore;

     {- Cross-Over Rate -}
          if (Random <= CrossOver_Rate) then
          begin
               SplitPos := Round(Random * Population[NewOrder[i]].NumberOfValues);
               {First Half from A}
               for j := 0 to SplitPos - 1 do
                   NewPopulation[i].Values[j] := SelectionA[j];
               {Second Half from B}
               for j := SplitPos to Population[NewOrder[i]].NumberOfValues - 1 do
                   NewPopulation[i].Values[j] := SelectionB[j];
          end
          else
          begin
               if (AFitness > BFitness) then
                  for j := 0 to NewPopulation[i].NumberOfValues - 1 do
                      NewPopulation[i].Values[j] := SelectionA[j]
               else
                   for j := 0 to NewPopulation[i].NumberOfValues - 1 do
                       NewPopulation[i].Values[j] := SelectionB[j];
          end;


     {- Mutation Rate -}
          for j := 0 to NewPopulation[i].NumberOfValues - 1 do
              if (Random <= Mutation_Rate) then
                 NewPopulation[i].Values[j] := (Random * 2) - 1;
     end;


     {- Update Chromosomes -}
     for i := 0 to NumberOfPopulation - 1 do
         for j := 0 to Population[i].NumberOfValues - 1 do
         begin
              Population[i].Values[j] := NewPopulation[i].Values[j];
              Population[i].FitnessScore := 0;
         end;

     {- Destroy Temporary Objects -}
     for i := 0 to NumberOfPopulation - 1 do
         NewPopulation[i].DeInit;
     NewPopulation := nil;
end;


{-- Chromosome Functions --}
constructor TChromosome.Init(Size: Integer);
begin
     NumberOfValues := Size;
     SetLength(Values, Size);
     FitnessScore := 0;
end;
destructor TChromosome.DeInit;
begin
     Values := Nil;
end;
procedure TChromosome.GetChromo(var Chromo: Array of Real);
var i: Integer;
begin
     if (Length(Chromo) = NumberOfValues) then
        for i := 0 to NumberOfValues - 1 do
            Chromo[i] := Values[i]
     else
     begin
          ShowMessage('ERROR <TChromosome.GetChromo()>: Non-Equal Array Sizes! tried<' +
                      IntToStr(Length(Chromo)) + '> actual<' + IntToStr(NumberOfValues) + '>');
          Halt;
     end;
end;
procedure TChromosome.PutChromo(Chromo: Array of Real);
var i: Integer;
begin
     if (Length(Chromo) = NumberOfValues) then
        for i := 0 to NumberOfValues - 1 do
            Values[i] := Chromo[i]
     else
     begin
          ShowMessage('ERROR <TChromosome.PutChromo>: Non-Equal Array Sizes! tried<' +
                      IntToStr(Length(Chromo)) + '> actual<' + IntToStr(NumberOfValues) + '>');
          Halt;
     end;
end;


{-- Neuron Functions --}
constructor TNeuron.Init(NumOfInputs: Integer; Digital: Boolean);
var i: Integer;
begin
     NumberOfInputs := NumOfInputs;
     DigitalOutput := Digital;

     {Set Array Size Of Weights}
     SetLength(Weight, NumOfInputs + 1);
     {Init Threshold}
     Weight[0] := (Random * 2) - 1;//Random - Random?
     {Init Weights}
     for i := 1 to NumOfInputs do
         Weight[i] := (Random * 2) - 1;//Random - Random?
end;
function TNeuron.Output(Inputs: Array of Real): Real;
var i: Integer;
    Activation: Real;
begin
     Activation := 0.0000000001;

     {Sum all weights * inputs}
     for i := 1 to NumberOfInputs do
         Activation := Activation + (Inputs[i - 1] * Weight[i]);

     {Test for Threshold}
     if (DigitalOutput) then
     begin
          if (Activation >= Weight[0]) then
             Result := 1
          else
              Result := 0;
     end
     else
     begin
          Activation := Activation + (Weight[0] * -1);
          {Apply Sigmoid}
          Result := Sigmoid(Activation, 1.0);
     end;
end;


{-- Neuron Layer Functions --}
constructor TNeuronLayer.Init(NumOfNeurons, NumOfInputsPerNeuron: Integer; Digital: Boolean);
var i: Integer;
begin
     NumberOfNeurons := NumOfNeurons;

     {Set Array Size Of Neurons}
     SetLength(Neurons, NumOfNeurons);
     for i := 0 to NumOfNeurons - 1 do
         Neurons[i] := TNeuron.Init(NumOfInputsPerNeuron, Digital);
end;
procedure TNeuronLayer.Execute(Inputs: Array of Real; var Outputs: Array of Real);
var i: Integer;
    Buffer: Array of Real;
begin
     SetLength(Buffer, NumberOfNeurons);
     for i := 0 to NumberOfNeurons - 1 do
         Buffer[i] := Neurons[i].Output(Inputs);
     for i := 0 to NumberOfNeurons - 1 do
         Outputs[i] := Buffer[i];
     Buffer := Nil;
end;


{-- Neural Net Functions --}
constructor TNeuralNet.Init(NumOfInputs: Integer; NeuronSet: Array of Integer; Digital: Boolean);
var i: Integer;
begin
     DigitalOutput := Digital;
     NumberOfInputs := NumOfInputs;
     NumberOfLayers := Length(NeuronSet);
     NumberOfOutputs := NeuronSet[NumberOfLayers - 1];

     SetLength(Layers, NumberOfLayers);

     Layers[0] := TNeuronLayer.Init(NeuronSet[0], NumberOfInputs, Digital);
     if (NumberOfLayers > 1) then
        for i := 1 to NumberOfLayers - 1 do
            Layers[i] := TNeuronLayer.Init(NeuronSet[i], Layers[i - 1].NumberOfNeurons, Digital);
end;
destructor TNeuralNet.Kill;
var i, j: Integer;
begin
     for i := 0 to NumberOfLayers - 1 do
     begin
          for j := 0 to Layers[i].NumberOfNeurons - 1 do
          begin
               Layers[i].Neurons[j].Weight := Nil;
               Layers[i].Neurons[j].Free;
          end;
          Layers[i].Neurons := Nil;
          Layers[i].Free;
     end;
     Layers := Nil;
end;
function TNeuralNet.GetNumberOfWeights: Integer;
var i, j: Integer;
begin
     Result := 0;
     if (NumberOfLayers > 0) then
        for i := 0 to NumberOfLayers - 1 do
            if (Layers[i].NumberOfNeurons > 0) then
               for j := 0 to Layers[i].NumberOfNeurons - 1 do
                   Result := Result + Layers[i].Neurons[j].NumberOfInputs + 1;
end;
procedure TNeuralNet.GetWeights(var Weights: Array of Real);
var i, j, k, w: Integer;
begin
//GetNumberOfWeights
     w := 0;
     if (NumberOfLayers > 0) then
        for i := 0 to NumberOfLayers - 1 do
            if (Layers[i].NumberOfNeurons > 0) then
               for j := 0 to Layers[i].NumberOfNeurons - 1 do
                   for k := 0 to Layers[i].Neurons[j].NumberOfInputs do
                   begin
                        Weights[w] := Layers[i].Neurons[j].Weight[k];
                        inc(w);
                   end;
end;
procedure TNeuralNet.PutWeights(Weights: Array of Real);
var i, j, k, w: Integer;
begin
     w := 0;
     if (NumberOfLayers > 0) then
        for i := 0 to NumberOfLayers - 1 do
            if (Layers[i].NumberOfNeurons > 0) then
               for j := 0 to Layers[i].NumberOfNeurons - 1 do
                   for k := 0 to Layers[i].Neurons[j].NumberOfInputs do
                   begin
                        Layers[i].Neurons[j].Weight[k] := Weights[w];
                        inc(w);
                   end;
end;
procedure TNeuralNet.Execute(Input: Array of Real; var Output: Array of Real);
var i, j: Integer;
    inBuffer, outBuffer: Array of Real;
begin
     if (Length(Input) = NumberOfInputs) then
     begin
          SetLength(outBuffer, Layers[0].NumberOfNeurons);
          Layers[0].Execute(Input, outBuffer);
          if (NumberOfLayers > 1) then
             for i := 1 to NumberOfLayers - 1 do
             begin
                  SetLength(inBuffer, Layers[i - 1].NumberOfNeurons);
                  for j := 0 to Length(inBuffer) - 1 do
                      inBuffer[j] := outBuffer[j];

                  SetLength(outBuffer, Layers[i].NumberOfNeurons);
                  Layers[i].Execute(inBuffer, outBuffer);
             end;
          {Resolve Output}
          for i := 0 to NumberOfOutputs - 1 do
              if (DigitalOutput) then
              begin
                   if (outBuffer[i] > 0) then
                      Output[i] := 1
                   else
                       Output[i] := 0;
              end
              else
                  Output[i] := outBuffer[i];
     end;
     inBuffer := Nil;
     outBuffer := Nil;
end;

end.
