unit MainUnit;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, AIUnit,
  StdCtrls, DXDraws, DXClass, Math, Spin, Buttons, AdvancedMathUnit;

type
  TMain = class(TDXForm)
    StartBtn: TButton;
    WorldView: TDXDraw;
    Timer: TDXTimer;
    ObjectImages: TDXImageList;
    PauseBtn: TSpeedButton;
    WorldSpeedBox: TSpinEdit;
    Label1: TLabel;
    FPSLabel: TLabel;
    GroupBox1: TGroupBox;
    Label2: TLabel;
    Label3: TLabel;
    Label7: TLabel;
    NumberOfBugsBox: TSpinEdit;
    NumberOfFoodBox: TSpinEdit;
    CreatureLifetimeBox: TSpinEdit;
    Label8: TLabel;
    EvolveOnlyBox: TSpinEdit;
    ObservationGroup: TGroupBox;
    FitnessLabel: TLabel;
    Label4: TLabel;
    Label5: TLabel;
    Label6: TLabel;
    CreatureSelector: TSpinEdit;
    GenomeMap: TDXDraw;
    GetStrongestBtn: TButton;
    GetWeakestBtn: TButton;
    LeftImpulseMeter: TDXDraw;
    RightImpulseMeter: TDXDraw;
    ManipulationGroup: TGroupBox;
    FeedCreatureBtn: TButton;
    PunishCreatureBtn: TButton;
    VisualToggleGroup: TGroupBox;
    ShowStatsBox: TCheckBox;
    WorldViewEnabled: TCheckBox;
    RecordNNIOBox: TCheckBox;
    LimitWeightsBox: TCheckBox;
    Label9: TLabel;
    procedure StartBtnClick(Sender: TObject);
    procedure TimerTimer(Sender: TObject; LagCount: Integer);
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure CreatureSelectorChange(Sender: TObject);
    procedure WorldViewClick(Sender: TObject);
    procedure WorldViewMouseMove(Sender: TObject; Shift: TShiftState; X, Y: Integer);
    procedure PauseBtnClick(Sender: TObject);
    procedure GetStrongestBtnClick(Sender: TObject);
    procedure GetWeakestBtnClick(Sender: TObject);
    procedure WorldSpeedBoxChange(Sender: TObject);
    procedure ShowStatsBoxClick(Sender: TObject);
    procedure FeedCreatureBtnClick(Sender: TObject);
    procedure WorldViewEnabledClick(Sender: TObject);
    procedure PunishCreatureBtnClick(Sender: TObject);
  private
    { Private declarations }
  public
    { Public declarations }
  end;
  TCreature = class(TObject)
   private
   public
    X, Y,
    VelX, VelY,
    VecX, VecY,
    Rotation, Angle, Speed,
    OldAngle, OldSpeed: Real;
    constructor Init(oX, oY: Integer; oAngle, oSpeed: Real);
    procedure Move;
  end;
  TFood = Record
    X, Y: Real;
  end;

const
  MAX_Population = 200;
  MAX_Food = 100;

  Mutation_Rate = 0.1;
  Max_Rotation = 0.3;{0.3}

var
  {World Values}
  NNMapNumberOfLayers: Integer = 2;
  NNMap: Array of Integer;
  ChromoLength: Integer;

  NumberOfCreatures: Integer = MAX_Population;
  NN       : Array[0 .. MAX_Population - 1] of TNeuralNet;
  Creature : Array[0 .. MAX_Population - 1] of TCreature;
  Genome   : Array[0 .. MAX_Population - 1] of TChromosome;

  NumberOfFood: Integer = MAX_Food;
  Food : Array[0 .. MAX_Food - 1] of TFood;

  {Statistics}
  NumberOfGenerations: LongInt;
  TotalFitness: LongInt;
  AverageFitness: Real;

  {Real Stats}
  SelectedCreature : Integer;
  TargetedFood     : Integer;

  {Mouse Data}
  MouseX, MouseY: Integer;

  Main: TMain;

implementation

{$R *.DFM}

constructor TCreature.Init(oX, oY: Integer; oAngle, oSpeed: Real);
begin
     X := oX;
     Y := oY;
     Angle := DegreesToRadians(oAngle);
     Speed := oSpeed;
     Rotation := 0.16;{Random}
     VecX := -sin(Angle);    
     VecY := cos(Angle);
end;
procedure TCreature.Move;
begin
     if (Angle > Pi_Mul_2) then
        Angle := Angle - Pi_Mul_2;

     if (Angle < -Pi_Mul_2) then
        Angle := Angle + Pi_Mul_2;

     if (Speed > 2) then
        Speed := 2;

     X := X + (VecX * Speed);
     Y := Y + (VecY * Speed);
end;


{-- Log Functions --}

function LogExists: Boolean;
begin
     if (FileExists('NN.log')) then
        Result := True
     else
         Result := False;
end;
procedure CreateLog;
var FileStream: TextFile;
begin
     AssignFile(FileStream, 'NN.log');
     Rewrite(FileStream);
     WriteLN(FileStream, 'NN Log Created: ' + DateTimeToStr(Now));
     WriteLN(FileStream, 'IN0 = Closest Normalized Food Vector X');
     WriteLN(FileStream, 'IN1 = Closest Normalized Food Vector Y');
     WriteLN(FileStream, 'IN2 = Normalized Directional Vector X');
     WriteLN(FileStream, 'IN3 = Normalized Directional Vector Y');
     WriteLN(FileStream, 'OUT0 = Left Impulse');
     WriteLN(FileStream, 'OUT1 = Right Impulse');
     CloseFile(FileStream);
end;
procedure AddToLog(Entry: String);
var FileStream: TextFile;
begin
     AssignFile(FileStream, 'NN.log');
     Append(FileStream);
     WriteLN(FileStream, Entry);
     CloseFile(FileStream);
end;


{-- Initialization Functions --}

procedure InitCreatures;
var i: Integer;
begin
     for i := 0 to MAX_Population - 1 do
         Creature[i] := TCreature.Init(Round(Main.WorldView.SurfaceWidth * Random), Round(Main.WorldView.SurfaceHeight * Random), Random * 360, Random);
end;
procedure InitFood;
var i: Integer;
begin
     for i := 0 to MAX_Food - 1 do
     begin
          Food[i].X := Main.WorldView.SurfaceWidth * Random;
          Food[i].Y := Main.WorldView.SurfaceHeight * Random;
     end;
end;
procedure ReplaceFood(Index: Integer);
begin
     Food[Index].X := Main.WorldView.SurfaceWidth * Random;
     Food[Index].Y := Main.WorldView.SurfaceHeight * Random;
end;

procedure NextGeneration;
begin
     {Apply Genetic Algorithm}
     GetChromosomes(NN, Genome, ChromoLength);
     Apply_Basic_Algorithm(Genome, NumberOfCreatures, ChromoLength);
     PutChromosomes(NN, Genome, ChromoLength);

     {Update Statistics}
     inc(NumberOfGenerations);
     AverageFitness := TotalFitness / NumberOfCreatures;
     TotalFitness := 0;

     {Reset LifeTime Counter}
     Main.Timer.Tag := Main.CreatureLifetimeBox.Value;
end;


{-- Control Functions --}
procedure MoveObjects;
var i, j: Integer;
    Inputs, Outputs: Array of Real;
    ClosestDist, ClosestX, ClosestY: Real;
    NormClosestX, NormClosestY: Real;
    LeftImpulse, RightImpulse: Real;
    Rotation: Real;
begin
     SetLength(Inputs, 4);
     SetLength(Outputs, NNMap[Length(NNMap) - 1]);
     for i := 0 to NumberOfCreatures - 1 do
     begin
          {Find Closest Food to Creature}
          ClosestDist := GetDistance(Creature[i].X, Creature[i].Y, Food[0].X, Food[0].Y);
          ClosestX := Food[0].X;
          ClosestY := Food[0].Y;
          for j := 1 to NumberOfFood - 1 do
              if (GetDistance(Creature[i].X, Creature[i].Y, Food[j].X, Food[j].Y) < ClosestDist) then
              begin
                   ClosestDist := GetDistance(Creature[i].X, Creature[i].Y, Food[j].X, Food[j].Y);
                   ClosestX := Food[j].X;
                   ClosestY := Food[j].Y;

                   if (i = SelectedCreature) then
                      {Assign Food Target for Creature Selector}
                      TargetedFood := j;
              end;
          {Normalize Closest Food Vector}
          NormClosestX := ClosestX / sqrt(ClosestX * ClosestX + ClosestY * ClosestY);
          NormClosestY := ClosestY / sqrt(ClosestX * ClosestX + ClosestY * ClosestY);

          {-- Neural Net's Inputs --}
          {Closest X, Y}
          Inputs[0] := NormClosestX;
          Inputs[1] := NormClosestY;

          {Creature's X, Y}
          Inputs[2] := Creature[i].VecX;
          Inputs[3] := Creature[i].VecY;

          if (Inputs[0] > 1.0) or (Inputs[1] > 1.0) then
          begin
               AddToLog('Missed Food Vector Calculation for Bug[' + IntToStr(i) +
                      ']: IN0: ' + FloatToStrF(Inputs[0], ffFixed, 9, 8) +
                      ' IN1: ' + FloatToStrF(Inputs[1], ffFixed, 9, 8) +
                      ' ');
               for j := 0 to NumberOfFood - 1 do
                   AddToLog('Food[' + IntToStr(j) + '].X = ' + FloatToStrF(Food[i].X, ffFixed, 9, 8) +
                            'Food[' + IntToStr(j) + '].Y = ' + FloatToStrF(Food[i].Y, ffFixed, 9, 8));
          end;

          { <<----- --------------- ----->> }
          { <<--- Update AI Reactions --->> }
          { <<----- --------------- ----->> }
          NN[i].Execute(Inputs, Outputs);

          {Log Inputs and Outputs}
          if (Main.RecordNNIOBox.Checked) then
             AddToLog('Execute[' + IntToStr(i) + ']: IN0: ' + FloatToStrF(Inputs[0], ffFixed, 9, 8) +
                      ' IN1: ' + FloatToStrF(Inputs[1], ffFixed, 9, 8) +
                      ' IN2: ' + FloatToStrF(Inputs[2], ffFixed, 9, 8) +
                      ' IN3: ' + FloatToStrF(Inputs[3], ffFixed, 9, 8) +
                      ' OUT0: ' + FloatToStrF(Outputs[0], ffFixed, 9, 8) +
                      ' OUT1: ' + FloatToStrF(Outputs[1], ffFixed, 9, 8));

          {Adjust Speed and Angle from Creature's Neural Net Outputs}
          LeftImpulse := Outputs[0];
          RightImpulse := Outputs[1];

          Rotation := LeftImpulse - RightImpulse;
          if (Rotation > Max_Rotation) then
             Rotation := Max_Rotation;
          if (Rotation < -Max_Rotation) then
             Rotation := -Max_Rotation;

          Creature[i].Angle := Creature[i].Angle + Rotation;
          Creature[i].Speed := LeftImpulse + RightImpulse;

          Creature[i].VecX := -sin(Creature[i].Angle);
          Creature[i].VecY := cos(Creature[i].Angle);

          {Execute Movement}
          Creature[i].Move;

          {Left or Right Border}
          if (Creature[i].X < 0) then
             Creature[i].X := Main.WorldView.SurfaceWidth;
          if (Creature[i].X > Main.WorldView.SurfaceWidth) then
             Creature[i].X := 0;
          {Top or Bottom Border}
          if (Creature[i].Y < 0) then
             Creature[i].Y := Main.WorldView.SurfaceHeight;
          if (Creature[i].Y > Main.WorldView.SurfaceHeight) then
             Creature[i].Y := 0;

          {Draw Left and Right Impulse Meters for Selected Bug}
          if (Main.ShowStatsBox.Checked) then
             if (i = SelectedCreature) then
             begin
                  with Main.LeftImpulseMeter do
                  begin
                       Hint := FloatToStrF(LeftImpulse, ffFixed, 5, 4);
                       Surface.Fill(clBlack);
                       Surface.Canvas.Pen.Color := clLime;
                       Surface.Canvas.Brush.Color := clGreen;
                       Surface.Canvas.Rectangle(0, SurfaceHeight - Round(SurfaceHeight * LeftImpulse), SurfaceWidth, SurfaceHeight);
                       Surface.Canvas.Release;
                       Surface.Canvas.Font.Color := clYellow;
                       Surface.Canvas.Brush.Style := bsClear;
                       Surface.Canvas.TextOut(1, 1, 'L');
                       Surface.Canvas.Release;
                       Flip;
                  end;
                  with Main.RightImpulseMeter do
                  begin
                       Hint := FloatToStrF(RightImpulse, ffFixed, 5, 4);
                       Surface.Fill(clBlack);
                       Surface.Canvas.Pen.Color := clRed;
                       Surface.Canvas.Brush.Color := clMaroon;
                       Surface.Canvas.Rectangle(0, SurfaceHeight - Round(SurfaceHeight * RightImpulse), SurfaceWidth, SurfaceHeight);
                       Surface.Canvas.Release;
                       Surface.Canvas.Font.Color := clYellow;
                       Surface.Canvas.Brush.Style := bsClear;
                       Surface.Canvas.TextOut(1, 1, 'R');
                       Surface.Canvas.Release;
                       Flip;
                  end;
             end;

          {-V- Complex Border Method -V-}
          {Left or Right Border}
{          if (Creature[i].X < 0) or
             (Creature[i].X > Main.WorldView.SurfaceWidth) then
          begin
               Creature[i].VelX := Creature[i].VelX * -1;
               Creature[i].Angle := ArcCos(Creature[i].VelX / Creature[i].Speed) / Pi_Div_180 + 90;
          end;}
          {Top or Bottom Border}
{          if (Creature[i].Y < 0) or
             (Creature[i].Y > Main.WorldView.SurfaceHeight) then
          begin
               Creature[i].VelY := Creature[i].VelY * -1;
               Creature[i].Angle := ArcSin(Creature[i].VelY / Creature[i].Speed) / Pi_Div_180 + 90;
          end;}
     end;
    Inputs := Nil;
    Outputs := Nil;
end;
procedure EatFood;
var i, j: Integer;
begin
     for i := 0 to NumberOfCreatures - 1 do
         for j := 0 to NumberOfFood - 1 do
             if (GetDistance(Creature[i].X, Creature[i].Y, Food[j].X, Food[j].Y) < 6) then
             begin
                  NN[i].FitnessScore := NN[i].FitnessScore + 1;
                  inc(TotalFitness);
                  ReplaceFood(j);
             end;
end;


{-- Graphical Functions --}
procedure OutlinedTextOut(Canvas: TDirectDrawSurfaceCanvas; X, Y: Integer; TextColor, OutlineColor: TColor; Text: String);
begin
     Canvas.Font.Color := OutlineColor;
     Canvas.TextOut(X - 1, Y, Text);
     Canvas.TextOut(X + 1, Y, Text);
     Canvas.TextOut(X, Y - 1, Text);
     Canvas.TextOut(X, Y + 1, Text);
     Canvas.Font.Color := TextColor;
     Canvas.TextOut(X, Y, Text);
end;
procedure ClearWorldObjects;
begin
     with Main.WorldView do
     begin
          {World Area}
          Surface.Canvas.Brush.Color := clBlack;
          Surface.Canvas.Pen.Color := clMaroon;
          Surface.Canvas.Rectangle(0, 0, SurfaceWidth, SurfaceHeight);
          Surface.Canvas.Release;
          Flip;
     end;
end;
procedure DrawWorldObjects;
begin
     with Main.WorldView do
     begin
          {World Area}
          Surface.Canvas.Brush.Color := clNone;
          Surface.Canvas.Pen.Color := clRed;
          Surface.Canvas.Rectangle(0, 0, SurfaceWidth, SurfaceHeight);
          Surface.Canvas.Release;
     end;
end;
procedure DrawStats;
begin
     with Main.WorldView do
     begin
          Surface.Canvas.Brush.Style := bsClear;
//          OutlinedTextOut(Surface.Canvas, SurfaceWidth - 50, 1, clYellow, $002222, 'Bug #' + IntToStr(SelectedCreature));
          OutlinedTextOut(Surface.Canvas, 1, 1, clYellow, $002222,
                          'Generation ' + IntToStr(NumberOfGenerations));
          OutlinedTextOut(Surface.Canvas, 1, 14, clYellow, $002222,
                          'Time To Live ' + IntToStr(Main.Timer.Tag));
          OutlinedTextOut(Surface.Canvas, 1, SurfaceHeight - 14, clYellow, $002222,
                          'Average Fitness of Last Generation: ' +
                          FloatToStrF(AverageFitness, ffFixed, 2, 2));
          OutlinedTextOut(Surface.Canvas, 1, SurfaceHeight - 28, clYellow, $002222,
                          'Total Fitness: ' + IntToStr(TotalFitness));
          Surface.Canvas.Release;
     end;
end;
procedure ClearCreatureData;
begin
     with Main.GenomeMap do
     begin
          Surface.Fill(clBlack);
          Surface.Canvas.Pen.Color := $102020;
          Surface.Canvas.MoveTo(Round((1 / ChromoLength) * SurfaceWidth),
                                SurfaceHeight div 2);
          Surface.Canvas.LineTo(Round((ChromoLength / (ChromoLength + 1)) * SurfaceWidth),
                                SurfaceHeight div 2);
          Surface.Canvas.Release;
          Flip;
     end;
     with Main.LeftImpulseMeter do
     begin
          Hint := '';
          Surface.Fill(clBlack);
          Surface.Canvas.Font.Color := clOlive;
          Surface.Canvas.TextOut(1, 1, 'L');
          Surface.Canvas.Release;
          Flip;
     end;
     with Main.RightImpulseMeter do
     begin
          Hint := '';
          Surface.Fill(clBlack);
          Surface.Canvas.Font.Color := clOlive;
          Surface.Canvas.TextOut(1, 1, 'R');
          Surface.Canvas.Release;
          Flip;
     end;
end;
procedure DisplayCreatureData;
var i: Integer;
    Weights: Array of Real;
begin
     SetLength(Weights, ChromoLength);
     NN[SelectedCreature].GetWeights(Weights);

     Main.FitnessLabel.Caption := 'Fitness = ' + IntToStr(NN[SelectedCreature].FitnessScore);
     with Main.GenomeMap do
     begin
          Surface.Fill(clBlack);
          Surface.Canvas.Pen.Color := $102020;
          Surface.Canvas.MoveTo(Round((1 / ChromoLength) * SurfaceWidth),
                                SurfaceHeight div 2);
          Surface.Canvas.LineTo(Round((ChromoLength / (ChromoLength + 1)) * SurfaceWidth),
                                SurfaceHeight div 2);
          Surface.Canvas.Pen.Color := clYellow;
          for i := 1 to ChromoLength do
          begin
               Surface.Canvas.MoveTo(Round((i / (ChromoLength + 1)) * SurfaceWidth) - 1,
                                     SurfaceHeight div 2);
               Surface.Canvas.LineTo(Round((i / (ChromoLength + 1)) * SurfaceWidth) - 1,
                                     Round((SurfaceHeight / 2) - (Weights[i - 1] * (SurfaceHeight / 2))));
               Surface.Canvas.Release;
          end;
          Flip;
     end;
end;
procedure DrawCreatures;
var i, x, y: Integer;
begin
     for i := 0 to NumberOfCreatures - 1 do
     with Main.ObjectImages do
     begin
          x := Round(Creature[i].X);
          y := Round(Creature[i].Y);

          Items[0].DrawRotate(Main.WorldView.Surface, x, y, 9, 9, 0, 0.5, 0.5, DegreesTo256(Round(RadToDeg(Creature[i].Angle) - 180)));
          if (i = SelectedCreature) then
          begin
               try
                 Main.WorldView.Surface.Canvas.Brush.Style := bsClear;

                 if (Main.Timer.Tag mod 2 = 0) or (Main.Timer.Tag mod 3 = 0) then
                    Main.WorldView.Surface.Canvas.Pen.Color := $ff0000
                 else
                     Main.WorldView.Surface.Canvas.Pen.Color := $ff5555;

                    Main.WorldView.Surface.Canvas.Ellipse(x - 7, y - 7, x + 7, y + 7);
                    Main.WorldView.Surface.Canvas.Ellipse(x - 8, y - 8, x + 8, y + 8);
               finally
                 Main.WorldView.Surface.Canvas.Release;
               end;
          end;
     end;
end;
procedure DrawFood;
var i, x, y: Integer;
begin
     for i := 0 to NumberOfFood - 1 do
     with Main.ObjectImages do
     begin
          x := Round(Food[i].X - 4.5);
          y := Round(Food[i].Y - 4.5);

          Items[1].Draw(Main.WorldView.Surface, x, y, 0);
          if (i = TargetedFood) then
          begin
               Main.WorldView.Surface.Canvas.Brush.Style := bsClear;
               Main.WorldView.Surface.Canvas.Pen.Color := $551155;

               if (Main.Timer.Tag mod 2 = 0) then
                  Main.WorldView.Surface.Canvas.Ellipse(x - 1, y - 1, x + 10, y + 10);
               if (Main.Timer.Tag mod 3 = 0) then
                  Main.WorldView.Surface.Canvas.Ellipse(x - 2, y - 2, x + 11, y + 11)
               else
                   Main.WorldView.Surface.Canvas.Ellipse(x - 3, y - 3, x + 12, y + 12);
               Main.WorldView.Surface.Canvas.Release;
          end;
     end;
end;

procedure TMain.StartBtnClick(Sender: TObject);
var i: Integer;
begin
     {Update Visual Controls}
     StartBtn.Caption := 'Restart';
     StartBtn.Tag := 1;
     CreatureSelector.Enabled := True;
     GetStrongestBtn.Enabled := True;
     GetWeakestBtn.Enabled := True;
     FeedCreatureBtn.Enabled := True;
     PunishCreatureBtn.Enabled := True;

     {Init Neural Net}
     SetLength(NNMap, 2);
     NNMap[0] := 6; // Hidden Layer
     NNMap[1] := 2; // Output Layer

     for i := 0 to MAX_Population - 1 do
     begin
          NN[i] := TNeuralNet.Init(4, NNMap, False);

          ChromoLength := NN[i].GetNumberOfWeights;

          Genome[i] := TChromosome.Init(ChromoLength);
     end;

     InitCreatures;
     InitFood;

     NumberOfCreatures := NumberOfBugsBox.Value;
     NumberOfFood := NumberOfFoodBox.Value;

     {Init Stats}
     NumberOfGenerations := 0;
     TotalFitness := 0;
     AverageFitness := 0.0;

     {Init Timer & Creature LifeTime Counter}
     Timer.Tag := CreatureLifetimeBox.Value;
     Timer.Enabled := True;
end;

procedure TMain.FormCreate(Sender: TObject);
var Buffer, SeedBuffer: Word;
begin
     {Initalize WorldView}
     if (not WorldView.CanDraw) then
        Exit;
     WorldView.Surface.Fill(0);

     {Assign Random Seed with the MilliSecond of CPU clock}
     DecodeTime(Time, Buffer, Buffer, Buffer, SeedBuffer);
     RandSeed := SeedBuffer;
     Randomize;

     {Assign Selected Creature}
     SelectedCreature := 0;
     CreatureSelector.MaxValue := MAX_Population - 1;

     {Initalize Log}
     if (not LogExists) then
        CreateLog;

     AddToLog('Program Started: ' + DateTimeToStr(Now));
end;

procedure TMain.FormDestroy(Sender: TObject);
var i: Integer;
begin
     if (StartBtn.Caption = 'Restart') then
        for i := 0 to MAX_Population - 1 do
        begin
             NN[i].Kill;
             Genome[i].DeInit;
             Creature[i].Free;
        end;

     if (not LogExists) then
        CreateLog;

     AddToLog('Program Closed: ' + DateTimeToStr(Now));
end;

procedure TMain.CreatureSelectorChange(Sender: TObject);
begin
     if (CreatureSelector.Value > NumberOfCreatures - 1) then
     begin
          CreatureSelector.Value := CreatureSelector.Value - 1;
          CreatureSelector.Tag := 1;
     end;

     if (CreatureSelector.Tag = 0) then
        SelectedCreature := CreatureSelector.Value
     else
         CreatureSelector.Tag := 0;
end;

procedure TMain.WorldViewClick(Sender: TObject);
var i: Integer;
    ClosestDistance: Real;
    ClosestCreature: Integer;
begin
     if (StartBtn.Tag = 1) then
     if (NumberOfCreatures > 0) then
     begin
          ClosestCreature := 0;
          ClosestDistance := GetDistance(MouseX, MouseY, Creature[0].X, Creature[0].Y);
          for i := 1 to NumberOfCreatures - 1 do
              if (GetDistance(MouseX, MouseY, Creature[i].X, Creature[i].Y) < ClosestDistance) then
              begin
                   ClosestCreature := i;
                   ClosestDistance := GetDistance(MouseX, MouseY, Creature[i].X, Creature[i].Y);
              end;
          SelectedCreature := ClosestCreature;
          CreatureSelector.Value := SelectedCreature;
     end;
end;

procedure TMain.WorldViewMouseMove(Sender: TObject; Shift: TShiftState; X,
  Y: Integer);
begin
     MouseX := X;
     MouseY := Y;
end;

procedure TMain.PauseBtnClick(Sender: TObject);
begin
     if (Timer.Enabled) then
     begin
          Timer.Enabled := False;
          PauseBtn.Hint := 'Unpause';
     end
     else
     begin
          Timer.Enabled := True;
          PauseBtn.Hint := 'Pause';
     end;
end;

procedure TMain.WorldSpeedBoxChange(Sender: TObject);
begin
     Timer.Interval := WorldSpeedBox.Value * 10;
end;

procedure TMain.GetStrongestBtnClick(Sender: TObject);
var i: Integer;
    StrongestCreature,
    StrongestFitness: Integer;
begin
     if (NumberOfCreatures > 0) then
     begin
          StrongestCreature := 0;
          StrongestFitness := NN[StrongestCreature].FitnessScore;
          for i := 1 to NumberOfCreatures - 1 do
              if (NN[i].FitnessScore > StrongestFitness) then
              begin
                   StrongestCreature := i;
                   StrongestFitness := NN[StrongestCreature].FitnessScore;
              end;
          SelectedCreature := StrongestCreature;
          CreatureSelector.Value := SelectedCreature;
     end;
end;
procedure TMain.GetWeakestBtnClick(Sender: TObject);
var i: Integer;
    WeakestCreature,
    WeakestFitness: Integer;
begin
     if (NumberOfCreatures > 0) then
     begin
          WeakestCreature := 0;
          WeakestFitness := NN[WeakestCreature].FitnessScore;
          for i := 1 to NumberOfCreatures - 1 do
              if (NN[i].FitnessScore < WeakestFitness) then
              begin
                   WeakestCreature := i;
                   WeakestFitness := NN[WeakestCreature].FitnessScore;
              end;
          SelectedCreature := WeakestCreature;
          CreatureSelector.Value := SelectedCreature;
     end;
end;

procedure TMain.ShowStatsBoxClick(Sender: TObject);
begin
     if (not ShowStatsBox.Checked) then
        ClearCreatureData;
end;

procedure TMain.FeedCreatureBtnClick(Sender: TObject);
begin
     inc(NN[SelectedCreature].FitnessScore);
end;
procedure TMain.PunishCreatureBtnClick(Sender: TObject);
begin
     dec(NN[SelectedCreature].FitnessScore);
end;

procedure TMain.WorldViewEnabledClick(Sender: TObject);
begin
     if (not WorldViewEnabled.Checked) then
        ClearWorldObjects;
end;

procedure TMain.TimerTimer(Sender: TObject; LagCount: Integer);
begin
     {Frames Per Second}
     FPSLabel.Caption := 'FPS = ' + IntToStr(Timer.FrameRate) + '(' + IntToStr(LagCount) +
                         ' Lag)';

     {World Timer}
     Timer.Tag := Timer.Tag - 1;
     if (Timer.Tag = 0) then
        NextGeneration;

     {Movvement}
     MoveObjects;

     {Food Consumption}
     EatFood;

     {Graphics}
     if (WorldView.CanDraw) then
     begin
          {Draw World View Items}
          if (WorldViewEnabled.Checked) then
          begin
               WorldView.Surface.Fill(clBlack);
               DrawWorldObjects;
               DrawFood;
               DrawCreatures;
               DrawStats;
               WorldView.Flip;
          end;
          {Draw GenomeMap}
          if (ShowStatsBox.Checked) then
             DisplayCreatureData;
     end;
end;

end.
